/*
Lerner 1.0
Copyright (c) 2011, Dave Hein
See end of file for terms of use
*/
/*
**** Uncomment this for Spin ****
obj
  ser : "cserial"
  i2c : "Basic_I2C_Driver"
*/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>
#include <sys/timeb.h>
#include <conio.h>
#include <ctype.h>

// Constants
#define TYPE_NOUN    1
#define TYPE_VERB    2
#define TYPE_PRONOUN 3
#define TYPE_ADJECT  4
#define TYPE_OTHER   5
#define TYPE_NAME    6
#define TYPE_ARTICLE 7
#define TYPE_QUERY   8

#define TYPE_RELATED 32
#define TYPE_EQUATE  33
#define TYPE_RELATE2 34

#define TYPE_MASK    0x3f
#define DELETE_MASK  0x40

#define MEMORY_MASK  31

#define BUF_SIZE 10000

#ifdef CSPIN_FLAG
#define TICKS_PER_SECOND clkfreq
#else
#define TICKS_PER_SECOND 1000
#define FILE_IO
#endif

void SomethingWrong(char *str);
void AddEquate(char *ptr1, char *ptr2);
int Match(char *ptr1, char *ptr2);
char *GetNextItem(char *ptr);
void PrintStatement(char *name, char *ptr1, int remember);
int GetPlural(char *name);

// Global variables
#ifdef CSPIN_FLAG
int seed = 1;
#endif
int debugmode = 0;
int changeflag = 0;
int MemoryIndex = 0;
char *types[] = {"noun", "verb", "pronoun", "adjective",
    "other", "name", "article", "query"};

// Basic vocabulary
int basicsize = 249;
#ifndef CSPIN_FLAG
char basicvocab[] = {
0x03,0x49,0x00,0x00,0x19,0x02,0x61,0x6d,0x00,0x00,0x6b,0x07,0x61,0x00,0x00,0x00,
0x01,0x68,0x75,0x6d,0x61,0x6e,0x00,0x00,0x00,0x20,0x00,0x05,0x00,0x10,0x00,0x00,
0x03,0x79,0x6f,0x75,0x00,0x00,0x38,0x02,0x61,0x72,0x65,0x00,0x00,0x70,0x06,0x4c,
0x65,0x72,0x6e,0x65,0x72,0x00,0x00,0x50,0x20,0x00,0x27,0x00,0x2e,0x00,0x61,0x02,
0x69,0x73,0x00,0x00,0x75,0x01,0x70,0x72,0x6f,0x67,0x72,0x61,0x6d,0x00,0x00,0x00,
0x20,0x00,0x3f,0x00,0x45,0x00,0x66,0x02,0x65,0x71,0x75,0x61,0x6c,0x73,0x00,0x01,
0x17,0x21,0x00,0x2e,0x00,0x00,0x21,0x00,0x20,0x00,0x00,0x21,0x00,0x27,0x00,0x7a,
0x21,0x00,0x05,0x01,0x04,0x21,0x00,0x05,0x00,0x7f,0x21,0x00,0x3f,0x00,0x00,0x21,
0x00,0x27,0x00,0x00,0x21,0x00,0x3f,0x00,0x00,0x02,0x65,0x71,0x75,0x61,0x6c,0x00,
0x01,0x12,0x21,0x00,0x57,0x00,0x00,0x21,0x01,0x09,0x00,0x00,0x08,0x77,0x68,0x6f,
0x00,0x01,0x2b,0x08,0x77,0x68,0x61,0x74,0x00,0x01,0x30,0x21,0x01,0x23,0x00,0x00,
0x21,0x01,0x1c,0x00,0x00,0x01,0x63,0x6f,0x6d,0x70,0x75,0x74,0x65,0x72,0x00,0x01,
0x49,0x02,0x72,0x75,0x6e,0x73,0x00,0x01,0x74,0x20,0x01,0x41,0x00,0x45,0x00,0x00,
0x02,0x64,0x6f,0x65,0x73,0x00,0x01,0x6a,0x02,0x72,0x75,0x6e,0x00,0x01,0x6f,0x02,
0x64,0x6f,0x00,0x01,0x65,0x21,0x01,0x50,0x00,0x00,0x21,0x01,0x5f,0x00,0x00,0x21,
0x01,0x41,0x00,0x00,0x21,0x01,0x58,0x00,0x00};
#else
char basicvocab[1];
#endif

// ****************************************************
// Replace the following lines with the dumped output
int version = 2;
int size = 0;
char vocab[BUF_SIZE];
char *Memory[MEMORY_MASK+1];
// ****************************************************

#ifdef CSPIN_FLAG
typedef struct timeb {
    int dummy;
} timebT;

// Generate a random number
int rand(void)
{
    return (?seed) & 0x7fffffff;
}

// Check if a key has been pressed
int kbhit(void)
{
    return ser_kbhit;
}

// Return the system tick count
int GetTicks(void)
{
    return cnt;
}

// Convert ASCII character to upper case
int toupper(int val)
{
    if (val >= 'a' && val <= 'z') val -= 'a' - 'A';
    return val;
}
#else
// Return the system tick count
int GetTicks(void)
{
    int seconds;
    struct timeb timebuf;

    ftime(&timebuf);
    seconds = timebuf.time;
    return (seconds * 1000) + timebuf.millitm;
}
#endif

// Find the next character that does not match val
char *SkipChar(char *str, int val)
{
    while (*str)
    {
	if (*str != val) break;
	str++;
    }
    return str;
}

// Find the next character that matches val
char *FindChar(char *str, int val)
{
    while (*str)
    {
	if (*str == val) break;
	str++;
    }
    return str;
}

// Convert the space-delimited string to a list of tokens
int tokenize(char *str, char **tokens)
{
    char *ptr;
    int num = 0;
    int len = strlen(str);
    int val = str[len-1];

    if (len == 0) return 0;

    if (val < 'A' || (val > 'Z' && val < 'a') || val > 'z')
    {
	str[len+1] = 0;
	str[len] = str[len-1];
	str[len-1] = ' ';
    }

    while (*str)
    {
	str = SkipChar(str, ' ');
	if (*str == 0) break;
	tokens[num++] = str;
	ptr = FindChar(str, ' ');
	if (*ptr) *ptr++ = 0;
	str = ptr;
    }
    return num;
}

// Find the word "str" in the vocab table
char *FindWord(char *str)
{
    char *ptr = vocab;

    while (ptr < vocab + size)
    {
	if ((*ptr & TYPE_MASK) == TYPE_RELATED)
	{
	    ptr += 7;
	}
	else if ((*ptr & TYPE_MASK) == TYPE_RELATE2)
	{
	    ptr += 9;
	}
	else if ((*ptr & TYPE_MASK) == TYPE_EQUATE)
	{
	    ptr += 5;
	}
	else
	{
	    if ((*ptr & DELETE_MASK) == 0)
	    {
		if(strcmp(str, ptr+1) == 0) return ptr;
	    }
	    ptr += strlen(ptr) + 3;
	}
    }
    return 0;
}

// Reset the vocab table
void ResetVocab(void)
{
    char buf[100];
    printf("< Ary you sure you want to forget everything?\n");
    printf("> ");
    gets(buf);
    if (toupper(buf[0]) == 'Y')
    {
        size = 0;
	changeflag = 1;
	printf("< Memory erased\n");
    }
    else
        printf("< Reset aborted\n");
}

// Reset to a basic vocab table
void ResetBasicVocab(void)
{
    char buf[100];
    printf("< Ary you sure you want to reset to a basic word list?\n");
    printf("> ");
    gets(buf);
    if (toupper(buf[0]) == 'Y')
    {
	size = basicsize;
	memcpy(vocab, basicvocab, size);
	changeflag = 1;
	printf("< Memory reset to basic words\n");
    }
    else
        printf("< Reset aborted\n");
}

// Dump out a Spin formatted version of the vocab table
void DumpHexWords(void)
{
    int i = 0;
    char *ptr = vocab;
    printf("  version long %d\n", version);
    printf("  size long %d\n", size);
    printf("  vocab byte\n");
    printf("  byte ");
    while (ptr < vocab + size)
    {
        printf("$%2.2x", *ptr++);
        if (ptr == vocab + size) printf("\n");
        else if ((i++)%16 == 15) printf("\n  byte ");
        else printf(",");
    }
    printf("  byte 0[%d]\n", BUF_SIZE - size);
}

// Dump an ASCII formatted version of the vocab table
void DumpAsciiWords(char *name, int num)
{
    int i = 0;
    char *ptr1;
    int index, index2, index3;
    char *ptr = vocab;

    if (num > 1)
    {
	ptr = FindWord(name);
	if (!ptr)
	{
	    printf("$ %s not found\n", name);
	    return;
	}
    }

    while (ptr < vocab + size)
    {
	if ((*ptr & TYPE_MASK) == TYPE_RELATED)
	{
	    ptr += 7;
	}
	else if ((*ptr & TYPE_MASK) == TYPE_RELATE2)
	{
	    ptr += 9;
	}
	else if ((*ptr & TYPE_MASK) == TYPE_EQUATE)
	{
	    ptr += 5;
	}
	else
	{
	    if ((*ptr & DELETE_MASK) == 0)
	    {
		printf("$ %s - %s", ptr+1, types[ptr[0]-1]);
		ptr1 = ptr + strlen(ptr) + 1;
                while (ptr1 < vocab + size && (ptr1[0] || ptr1[1]))
                {
	            index = (ptr1[0] << 7) | ptr1[1];
                    ptr1 = vocab + index;
		    if ((ptr1[0] & TYPE_MASK) == TYPE_RELATED)
		    {
			if ((ptr1[0] & DELETE_MASK) == 0)
			{
	                    index = (ptr1[1] << 7) | ptr1[2];
                            index2 = (ptr1[3] << 7) | ptr1[4];
                            printf(" (%s %s)", vocab+index+1, vocab+index2+1);
			}
                        ptr1 += 5;
		    }
		    else if ((ptr1[0] & TYPE_MASK) == TYPE_RELATE2)
		    {
			if ((ptr1[0] & DELETE_MASK) == 0)
			{
	                    index = (ptr1[1] << 7) | ptr1[2];
                            index2 = (ptr1[3] << 7) | ptr1[4];
                            index3 = (ptr1[5] << 7) | ptr1[6];
                            printf(" (%s %s %s)", vocab+index+1, vocab+index2+1, vocab+index3+1);
			}
                        ptr1 += 7;
		    }
		    else if ((ptr1[0] & TYPE_MASK) == TYPE_EQUATE)
		    {
			if ((ptr1[0] & DELETE_MASK) == 0)
			{
	                    index = (ptr1[1] << 7) | ptr1[2];
                            printf(" (=%s)", vocab+index+1);
			}
                        ptr1 += 3;
		    }
		    else
	                printf("$ Invalid type = %2.2x\n", ptr1[0]);
                }
		printf("\n");
		if (num > 1) break;
	    }
	    ptr += strlen(ptr) + 3;
	}
    }
}

// Add a word to the vocab table
void AddWord(int attrib, char *str)
{
    char *ptr = vocab+size;
    char *ptr1 = ptr;

    if (size + strlen(str) + 4 > BUF_SIZE)
    {
	printf("$ Out of memory\n");
	return;
    }

    if (debugmode) printf("AddWord %s - %s\n", str, types[attrib-1]);

    *ptr++ = attrib;
    strcpy(ptr, str);
    ptr += strlen(ptr) + 1;
    *ptr++ = 0;
    *ptr = 0;
    size += strlen(str) + 4;
    changeflag = 1;
}

// Determine the attribute of an unknown word, and add it to the vocab table
int UnknownWord(char **tokens, int num, int i)
{
    char *str = tokens[i];
    char buf[100];
    char *ptr;
    int caseflag = 0;

    if (i == 0)
    {
	if (str[0] >= 'A' && str[0] <= 'Z')
	{
	    caseflag = 1;
	    str[0] += 'a' - 'A';
	    if (FindWord(str)) return 0;
	    str[0] -= 'a' - 'A';
	    AddWord(TYPE_NAME, str);
	    return 0;
	}
	else
	{
	    if (num > 2)
	    {
		if (strcmp(tokens[1], "equal") == 0 ||
		    strcmp(tokens[1], "equals") == 0)
		{
		    ptr = FindWord(tokens[2]);
		    if (ptr)
		    {
			AddWord(*ptr, tokens[0]);
			return 0;
		    }
		}
		else
		{
		    AddWord(TYPE_ADJECT, tokens[0]);
		    return 0;
		}
	    }
	}
    }
    else
    {
	ptr = FindWord(tokens[i-1]);
	if (ptr == 0)
	{
	    SomethingWrong(tokens[i-1]);
	}
	else
	{
	    if (ptr[0] == TYPE_ARTICLE && (i == 1 || i == num - 1))
	    {
	        AddWord(TYPE_NOUN, str);
	        return 0;
	    }
	    else if (i < 3 && (ptr[0] == TYPE_NOUN || ptr[0] == TYPE_NAME))
	    {
		AddWord(TYPE_VERB, str);
		return 0;
	    }
	    else if ((i > 1 && ptr[0] == TYPE_VERB) || i > 2)
	    {
		if (str[0] >= 'A' && str[0] <= 'Z')
		{
		    AddWord(TYPE_NAME, str);
		    return 0;
		}
		else if (i < num - 1 || ptr[0] == TYPE_VERB)
		{
		    AddWord(TYPE_ADJECT, str);
	            return 0;
		}
	    }
	    if (i == num - 1 && num >= 4)
	    {
	        ptr = FindWord(tokens[i-2]);
		if (ptr && ptr[0] == TYPE_ARTICLE)
		{
		    AddWord(TYPE_NOUN, str);
		    return 0;
		}
		else
		{
		    AddWord(TYPE_ADJECT, str);
		    return 0;
		}
	    }
	}
    }

    printf("< I don't know %s\n", str);
    while(1)
    {
        printf("< Is it a noun, verb, pronoun, adjective, Name, Article, query or ignore?\n");
	printf("> ");
        gets(buf);
	if (caseflag && buf[0] != 'N' && strcmp(str, "I"))
	    str[0] += 'a' - 'A';
        if (buf[0] == 'n')
	{
            AddWord(TYPE_NOUN, str);
	    break;
	}
        else if(buf[0] == 'v')
	{
            AddWord(TYPE_VERB, str);
	    break;
	}
        else if(buf[0] == 'a')
	{
            AddWord(TYPE_ADJECT, str);
	    break;
	}
        else if(buf[0] == 'p')
	{
            AddWord(TYPE_PRONOUN, str);
	    break;
	}
        else if(buf[0] == 'N')
	{
            AddWord(TYPE_NAME, str);
	    break;
	}
        else if(buf[0] == 'A')
	{
            AddWord(TYPE_ARTICLE, str);
	    break;
	}
        else if(buf[0] == 'q')
	{
            AddWord(TYPE_QUERY, str);
	    break;
	}
        else if(buf[0] == 'i')
	{
	    return 1;
	}
	printf("I don't understand\n");
    }
    return 0;
}

// Delete a word and all references to it from the vocab table
void DeleteWord(char **tokens, int num)
{
    char *ptr;
    char buf[100];
    int index, index2;

    if (num < 2)
    {
	printf("< I don't know what to forget\n");
	return;
    }
    ptr = FindWord(tokens[1]);
    if (ptr == 0)
    {
	printf("< I don't know %s\n", tokens[1]);
	return;
    }
    while (1)
    {
        printf("< Do you want me to forget %s?\n", tokens[1]);
	printf("> ");
        gets(buf);
        if (buf[0] == 'y')
        {
	    index = ptr - vocab;
	    // Delete the word
	    *ptr |= DELETE_MASK;
	    // Delete the chain
	    while (ptr = GetNextItem(ptr))
	    {
		*ptr |= DELETE_MASK;
	    }
	    // Delete any reference to this word
	    ptr = vocab;
	    while (ptr < vocab + size)
	    {
		if ((*ptr & TYPE_MASK) == TYPE_EQUATE)
		{
		    if ((*ptr & DELETE_MASK) == 0)
		    {
			index2 = (ptr[1]<<7) | ptr[2];
			if (index == index2) *ptr |= DELETE_MASK;
		    }
		    ptr += 5;
		}
		else if ((*ptr & TYPE_MASK) == TYPE_RELATED)
		{
		    if ((*ptr & DELETE_MASK) == 0)
		    {
			index2 = (ptr[1]<<7) | ptr[2];
			if (index == index2) *ptr |= DELETE_MASK;
			index2 = (ptr[3]<<7) | ptr[4];
			if (index == index2) *ptr |= DELETE_MASK;
		    }
		    ptr += 7;
		}
		else if ((*ptr & TYPE_MASK) == TYPE_RELATE2)
		{
		    if ((*ptr & DELETE_MASK) == 0)
		    {
			index2 = (ptr[1]<<7) | ptr[2];
			if (index == index2) *ptr |= DELETE_MASK;
			index2 = (ptr[3]<<7) | ptr[4];
			if (index == index2) *ptr |= DELETE_MASK;
			index2 = (ptr[5]<<7) | ptr[6];
			if (index == index2) *ptr |= DELETE_MASK;
		    }
		    ptr += 9;
		}
		else
		{
		    ptr += strlen(ptr) + 3;
		}
	    }
	    changeflag = 1;
	    printf("< %s forgotten\n", tokens[1]);
	    return;
        }
        else if (buf[0] == 'n')
        {
	    printf("< %s is not forgotten\n", tokens[1]);
	    return;
        }
        else
	{
	    printf("< I don't understand\n");
	}
    }
}

// Could find a word that was expected.  Print a message.
void SomethingWrong(char *str)
{
    printf("Something's wrong\n");
    printf("I couldn't find %s\n", str);
}

// Equate the first word to the second word
void AddEquate(char *ptr1, char *ptr2)
{
    int index;
    char *ptr3 = ptr1 + strlen(ptr1) + 1;

    if (size + 5 > BUF_SIZE)
    {
	printf("< Out of memory\n");
	return;
    }

    while (ptr3 < vocab+size && (ptr3[0] || ptr3[1]))
    {
	index = (ptr3[0] << 7) | ptr3[1];
        ptr3 = vocab+index;
	if ((ptr3[0] & TYPE_MASK) == TYPE_EQUATE)
	{
	    if ((ptr3[0] & DELETE_MASK) == 0)
	    {
	        index = (ptr3[1] << 7) | ptr3[2];
	        if (ptr2 == vocab+index) return;
	    }
	    ptr3 += 3;
	}
	else if ((ptr3[0] & TYPE_MASK) == TYPE_RELATED)
	{
            ptr3 += 5;
	}
	else if ((ptr3[0] & TYPE_MASK) == TYPE_RELATE2)
	{
            ptr3 += 7;
	}
	else
	    printf("Unknown type = %2.2x\n", ptr3[0]);
    }
    ptr3[0] = size >> 7;
    ptr3[1] = size & 127;
    ptr3 = vocab+size;
    *ptr3++ = TYPE_EQUATE;
    index = ptr2 - vocab;
    ptr3[0] = index >> 7;
    ptr3[1] = index & 127;
    ptr3[2] = 0;
    ptr3[3] = 0;
    size += 5;
    changeflag = 1;
}

// Get the next item in a chain
char *GetNextItem(char *ptr)
{
    int index;

    if (ptr == 0) return 0;

    while (ptr < vocab + size)
    {
        if ((ptr[0] & TYPE_MASK) == TYPE_EQUATE) ptr += 3;
        else if ((ptr[0] & TYPE_MASK) == TYPE_RELATED) ptr += 5;
        else if ((ptr[0] & TYPE_MASK) == TYPE_RELATE2) ptr += 7;
        else ptr += strlen(ptr) + 1;
        index = (ptr[0] << 7) | ptr[1];
	if (index == 0) break;
	ptr = vocab + index;
	if ((*ptr & DELETE_MASK) == 0) return ptr;
    }

    return 0;
}

// Get the next equate item in a chain
char *GetNextEquate(char *ptr)
{
    while (ptr = GetNextItem(ptr))
    {
	if (*ptr == TYPE_EQUATE) return ptr;
    }
    return 0;
}

// Match two word entries and their equates in the vocab table
int Match(char *ptr1, char *ptr2)
{
    int index;
    char *ptr3;
    char *name1;
    char *name2;

    name1 = ptr1 + 1;
    while (1)
    {
	ptr3 = ptr2;
	name2 = ptr2 + 1;
	while (1)
	{
	    if (strcmp(name1, name2) == 0) return 1;
	    ptr3 = GetNextEquate(ptr3);
	    if (!ptr3) break;
	    index = (ptr3[1]<<7) | ptr3[2];
	    name2 = vocab+index+1;
	}
        ptr1 = GetNextEquate(ptr1);
        if (!ptr1) break;
        index = (ptr1[1]<<7) | ptr1[2];
        name1 = vocab+index+1;
    }
    return 0;
}

// Remove articles from a token list
int RemoveArticles(char **tokens, int num)
{
    int i, j;
    char *ptr;

    for (i = 0; i < num;)
    {
	ptr = FindWord(tokens[i]);
	if (*ptr == 0)
	{
	    printf("RemoveArticles: Can't find %s\n", tokens[i]);
	    i++;
	    continue;
	}
	if (*ptr == TYPE_ARTICLE)
	{
	    num--;
	    for (j = i; j < num; j++)
	    {
		tokens[j] = tokens[j+1];
	    }
	    continue;
	}
	i++;
    }
    return num;
}

// Locate exact matches of a word and a verb in a related field
char *FindWords(char *word1, char *word2)
{
    char *ptr1 = FindWord(word1);
    char *ptr2;
    int index;

    while (ptr1)
    {
	if (ptr1[0] == TYPE_RELATED)
	{
	    index = (ptr1[1]<<7) | ptr1[2];
	    ptr2 = vocab+index;
	    if (strcmp(ptr2+1,word2) == 0) return ptr1;
	}
	ptr1 = GetNextItem(ptr1);
    }
    return 0;
}

// Return true if val is a vowel
int isvowel(int val)
{
    val = toupper(val);
    if (val == 'A' || val == 'E' || val == 'I' || val == 'O' || val == 'U')
        return 1;
    return 0;
}

void AddToMemory(char *name, char *ptr1)
{
    char *ptr0;

    ptr0 = FindWord(name);
    if (ptr0 == 0 && name[0] >= 'A' && name[0] <= 'Z')
    {
	name[0] += 'a' - 'A';
        ptr0 = FindWord(name);
	name[0] -= 'a' - 'A';
	if (ptr0 == 0) return;
    }
    Memory[MemoryIndex] = ptr0;
    Memory[MemoryIndex+1] = ptr1;
    MemoryIndex = (MemoryIndex + 2) & MEMORY_MASK;
}

void PrintMemory(void)
{
    char *ptr0;
    char *ptr1;
    int index = MemoryIndex;

    while (1)
    {
	ptr0 = Memory[index];
	ptr1 = Memory[index+1];
	if (ptr0 && ptr1)
	{
	    if ((*ptr0 & DELETE_MASK) == 0 && (*ptr1 & DELETE_MASK) == 0)
	        PrintStatement(ptr0+1, ptr1, 0);
	}
	index = (index + 2) & MEMORY_MASK;
	if (index == MemoryIndex) break;
    }
    printf("$\n");
}

int CheckMemory(char *ptr0, char *ptr1)
{
    char *ptr0a;
    char *ptr1a;
    int index = MemoryIndex;

    while (1)
    {
	ptr0a = Memory[index];
	ptr1a = Memory[index+1];
	if (ptr0 == ptr0a && ptr1 == ptr1a) return 1;
	index = (index + 2) & MEMORY_MASK;
	if (index == MemoryIndex) break;
    }

    return 0;
}

// Determine whether the word is plural, singular or something else
int GetPlural(char *name)
{
    int len = strlen(name);
    char *ptr1 = FindWord(name);
    //printf("GetPlural %s\n", name);
    if (strcmp(name, "I") == 0) return -1;
    if (len < 2) return 0;
    if (strcmp(name, "you") == 0 || strcmp(name, "You") == 0) return 1;
    if (ptr1 && *ptr1 != TYPE_NOUN && *ptr1 != TYPE_ADJECT) return 0;
    if (name[len-1] == 's' && name[len-2] != 's') return 1;
    return 0;
}

char *GetVerb(int plural, char *ptr1)
{
    char *verbptr;
    char *isptr = FindWord("is");

    if (isptr && Match(isptr, ptr1))
    {
        if (plural == 1)
            verbptr = "are";
        else if (plural == -1)
            verbptr = "am";
        else
            verbptr = "is";
    }
    else
        verbptr = ptr1 + 1;

    //printf("GetVerb %d %s %s\n", plural, ptr1+1, verbptr);

    return verbptr;
}

// Print a statement formed from the name and the field pointed to by ptr1
// Substitue I for you, and You for I in the name
void PrintStatement(char *name, char *ptr1, int remember)
{
    char *ptr;
    char nullstr = 0;
    char *Astr = &nullstr;
    int index, index2, index3;
    char namestr[100];
    char *astr = &nullstr;
    int plural;
    char *verbptr;

    if (remember)
    {
        if (strcmp(name, "you") == 0) name = "I";
        else if (strcmp(name, "I") == 0) name = "you";
    }

    strcpy(namestr, name);
    plural = GetPlural(namestr);
    ptr = FindWord(namestr);
    if (ptr)
    {
	if (ptr[0] == TYPE_NOUN && plural == 0 &&
	    (ptr1[0] == TYPE_RELATED || ptr1[0] == TYPE_RELATE2))
        {
	    if (isvowel(namestr[0]))
	        Astr = "An ";
	    else
	        Astr = "A ";
        }
    }

    if (!Astr[0] && namestr[0] >= 'a' && namestr[0] <= 'z')
    {
	namestr[0] -= 'a' - 'A';
    }

    if (remember)
        printf("< ");
    else
        printf("$ ");

    if (ptr1[0] == TYPE_RELATED)
    {
	index = (ptr1[1] << 7) | ptr1[2];
	index2 = (ptr1[3] << 7) | ptr1[4];
	if (vocab[index2] == TYPE_NOUN && GetPlural(vocab+index2+1) == 0)
	{
	    if (isvowel(vocab[index2+1]))
	        astr = "an ";
	    else
	        astr = "a ";
	}

	verbptr = GetVerb(plural, vocab+index);

	printf("%s%s %s %s%s\n", Astr, namestr, verbptr, astr, vocab+index2+1);
    }
    else if (ptr1[0] == TYPE_RELATE2)
    {
	index = (ptr1[1] << 7) | ptr1[2];
	index2 = (ptr1[3] << 7) | ptr1[4];
	index3 = (ptr1[5] << 7) | ptr1[6];
	if (vocab[index3] == TYPE_NOUN && GetPlural(vocab+index3+1) == 0)
	{
	    if (isvowel(vocab[index2+1]))
	        astr = "an ";
	    else
	        astr = "a ";
	}

	verbptr = GetVerb(plural, vocab+index);

	printf("%s%s %s %s%s %s\n", Astr, namestr, verbptr, astr, vocab+index2+1, vocab+index3+1);
    }
    else if (ptr1[0] == TYPE_EQUATE)
    {
	index = (ptr1[1] << 7) | ptr1[2];
	printf("%s is %s\n", namestr, vocab+index+1);
    }
    else
    {
	printf("%s is a %s\n", namestr, types[ptr1[0]-1]);
    }

    if (remember) AddToMemory(namestr, ptr1);
}

void PrintQuestion(char *ptr1)
{
    char nullstr = 0;
    char *Astr = &nullstr;
    char *isptr = FindWord("is");
    int plural = GetPlural(ptr1+1);
    char *verbptr = GetVerb(plural, isptr);

    if (ptr1[0] == TYPE_NOUN && plural != 1)
    {
        if (isvowel(ptr1[1]))
            Astr = "an ";
        else
            Astr = "a ";
    }

    printf("< What %s %s%s?\n", verbptr, Astr, ptr1+1);
}

// Process a list of tokens received from the user
void Process(char **tokens, int num, int question)
{
    int i, index, index2, index3;
    char *ptr[4];
    char *ptr1;
    char *ptr0;
    char *ptrx;
    int count;
    int counta;
    char *ptry;
    char *ptrya;
    int match = 0;
    int equalflag = 0;
    int tempflag = 0;

    // Equate the two words if it's an equal statement
    if (num == 3)
    {
        if (!question && (strcmp(tokens[1], "equal") == 0
            || strcmp(tokens[1], "equals") == 0))
        {
	    if (debugmode) printf("equals mode\n");
	    equalflag = 1;
	    ptr0 = FindWord(tokens[0]);
	    ptr1 = FindWord(tokens[2]);
	    if (!ptr0)
	    {
		SomethingWrong(tokens[0]);
		return;
	    }
	    if (!ptr1)
	    {
		SomethingWrong(tokens[2]);
		return;
	    }
	    AddEquate(ptr0, ptr1);
	    AddEquate(ptr1, ptr0);
	    return;
        }
    }

    // Check if the first word is do/does/did.
    // If so, remove the first word and make it a question.
    if (ptr0 = FindWord("do"))
    {
	if (debugmode) printf("does query\n");
        ptr1 = FindWord(tokens[0]);
	if (!ptr1)
	{
	    SomethingWrong(tokens[0]);
	    return;
	}
	if (Match(ptr0, ptr1))
	{
	    question = 1;
	    num--;
	    for (i = 0; i < num; i++)
	    {
		tokens[i] = tokens[i+1];
	    }
	}
    }

#if 0
    // Remove the articles if there are more than three items in the list
    if (num > 3) num = RemoveArticles(tokens, num);
#else
    // Remove the articles
    num = RemoveArticles(tokens, num);
#endif

    // There must be three items in the list or it is too short or too long
    if (num < 3)
    {
	printf("I don't understand\n");
	printf("Use more words\n");
	return;
    }
    if (num > 4)
    {
	printf("I don't understand\n");
	printf("Use less words\n");
	return;
    }

    // Locate the tokens in the vocab table
    for (i = 0; i < num; i++)
    {
        ptr[i] = FindWord(tokens[i]);
        if (ptr[i] == 0)
        {
	    SomethingWrong(tokens[i]);
	    return;
        }
    }

    // Change the word type if commanded to do so
    if (num == 3 && strcmp(tokens[1], "is") == 0)
    {
	if (strcmp(tokens[2], "noun") == 0)
	{
	    ptr[0][0] = TYPE_NOUN;
	    return;
	}
	else if (strcmp(tokens[2], "name") == 0)
	{
	    ptr[0][0] = TYPE_NAME;
	    return;
	}
	else if (strcmp(tokens[2], "adjective") == 0)
	{
	    ptr[0][0] = TYPE_ADJECT;
	    return;
	}
	else if (strcmp(tokens[2], "verb") == 0)
	{
	    ptr[0][0] = TYPE_VERB;
	    return;
	}
    }

    // If the first word is a query, the second word must be is/am/are
    if (ptr[0][0] == TYPE_QUERY)
    {
	if (debugmode) printf("query mode\n");
	question = 1;
	ptr1 = FindWord("is");
	tempflag = 0;
	//if (num != 3 || !ptr1 || !Match(ptr1, ptr[1]))
	if (num == 3)
	{
	    if (ptr1)
	    {
		if (Match(ptr1, ptr[1])) tempflag = 1;
	    }
	}
	if (!tempflag)
	{
	    printf("I don't understand\n");
	    return;
	}

        // Scan related entries and equate's related entries
        ptr1 = ptr[2];
        ptr0 = ptr1;
        ptrx = ptr1;
	ptry = ptr1;
	count = 0;
	// variables for items not in short term memory
	ptrya = 0;
	counta = 0;
	if (debugmode) PrintStatement(tokens[2], ptr1, 0);
        while (1)
        {
	    while (1)
	    {
	        ptr1 = GetNextItem(ptr1);
	        if (!ptr1) break;
	        if (ptr1[0] != TYPE_RELATED && ptr1[0] != TYPE_RELATE2) continue;
                //index = (ptr1[1] << 7) | ptr1[2];
                //index2 = (ptr1[3] << 7) | ptr1[4];
	        if (debugmode) PrintStatement(tokens[2], ptr1, 0);
		if (CheckMemory(ptr[2], ptr1) == 0)
		{
	            counta++;
	            if ((rand() % counta) == 0) ptrya = ptr1;
		}
		if (ptrya == 0)
		{
	            count++;
	            if ((rand() % count) == 0) ptry = ptr1;
		}
	    }
	    if (ptrya == 0) ptrya = ptry;
	    ptr0 = GetNextEquate(ptr0);
	    if (!ptr0) break;
            index = (ptr0[1] << 7) | ptr0[2];
	    ptrx = vocab+index;
	    ptr1 = ptrx;
        }
	if (ptrya)
	{
	    if (ptrya[0] == TYPE_RELATED ||
	        ptrya[0] == TYPE_RELATE2 || ptrya[1] == TYPE_EQUATE)
	        PrintStatement(tokens[2], ptrya, 1);
	    else
	        printf("< I don't know\n");
	}
	return;
    }

    // If not an equal statement or a question check if first word is a verb
    // Is so, swap the first and second tokens
    if (!equalflag && !question && ptr[0][0] == TYPE_VERB)
    {
	if (debugmode) printf("Lead verb query\n");
	question = 1;
	ptr1 = ptr[0];
	ptr[0] = ptr[1];
	ptr[1] = ptr1;
	ptr1 = tokens[0];
	tokens[0] = tokens[1];
	tokens[1] = ptr1;
    }

    // Look for a match to the three words.  If it is a question and
    // a match is found print "Yes"
    ptr1 = ptr[0];
    ptr0 = ptr1;
    ptrx = ptr1;
    while (1)
    {
	while (1)
	{
	    ptr1 = GetNextItem(ptr1);
	    if (!ptr1) break;
	    if (ptr1[0] != TYPE_RELATED && ptr1[0] != TYPE_RELATE2) continue;
            index = (ptr1[1] << 7) | ptr1[2];
            index2 = (ptr1[3] << 7) | ptr1[4];
	    tempflag = Match(ptr[1], vocab+index) && Match(ptr[2], vocab+index2);
	    if (num == 4)
	    {
                index3 = (ptr1[5] << 7) | ptr1[6];
		tempflag = tempflag && Match(ptr[3], vocab+index3);
	        if (debugmode) printf("%s (%s,%s) (%s,%s) (%s,%s)\n", ptrx+1,ptr[1]+1,
		    vocab+index+1,ptr[2]+1,vocab+index2+1,ptr[3],vocab+index3+1);
	    }
	    else
	    {
	        if (debugmode) printf("%s (%s,%s) (%s,%s)\n",
		    ptrx+1,ptr[1]+1,vocab+index+1,ptr[2]+1,vocab+index2+1);
	    }
            if (tempflag)
            {
	        if (!question)
	            printf("< I know that\n");
		else
                    printf("< Yes\n");
	        return;
            }
	}
	ptr0 = GetNextEquate(ptr0);
	if (!ptr0) break;
        index = (ptr0[1] << 7) | ptr0[2];
	ptrx = vocab+index;
	ptr1 = ptrx;
    }

    // Print "No" if it was a question and no match found
    if (question)
    {
	printf("< No\n");
	return;
    }

    // Add relationship if there was no match and it's not a question
    if (!match)
    {
        if (size + 1 + (num << 1) > BUF_SIZE)
        {
	    printf("< Out of memory\n");
	    return;
        }
	ptr1 = ptr[0];
	while (ptr0 = GetNextItem(ptr1)) ptr1 = ptr0;
	if (ptr1[0] == TYPE_RELATED) ptr1 += 5;
	else if (ptr1[0] == TYPE_RELATE2) ptr1 += 7;
	else if (ptr1[0] == TYPE_EQUATE) ptr1 += 3;
	else ptr1 += strlen(ptr1) + 1;
	ptr1[0] = size >> 7;
	ptr1[1] = size & 127;
	ptr1 = vocab+size;
	if (num == 3)
	    ptr1[0] = TYPE_RELATED;
	else
	    ptr1[0] = TYPE_RELATE2;
	index = ptr[1] - vocab;
	ptr1[1] = index >> 7;
	ptr1[2] = index & 127;
	index = ptr[2] - vocab;
	ptr1[3] = index >> 7;
	ptr1[4] = index & 127;
	if (num == 3)
	{
	    ptr1[5] = 0;
	    ptr1[6] = 0;
	    size += 7;
	}
	else
	{
	    index = ptr[3] - vocab;
	    ptr1[5] = index >> 7;
	    ptr1[6] = index & 127;
	    ptr1[7] = 0;
	    ptr1[8] = 0;
	    size += 9;
	}
	changeflag = 1;

	// Check if predicate is undefined
	index = 0;
	ptr1 = ptr[2];
	if (ptr1[0] < 32)
	{
	    ptr0 = ptr1 + strlen(ptr1) + 1;
	    index = (ptr0[0]<<7) | ptr0[1];
	}

	if (index == 0) PrintQuestion(ptr1);
	else printf("< I will remember that\n");
    }
}

// Wait from 3 to 10 seconds for input.  Print a random fact from
// the vocab table if no input received.
void WaitForInput(void)
{
    int waittime = ((rand() % 8) + 3) * TICKS_PER_SECOND;
    int starttime = GetTicks();
    char *ptr1;
    char *ptr2;
    char *ptr1a;
    char *ptr2a;
    int index, count;
    int FirstTime;

    printf("> ");
#ifndef CSPIN_FLAG
    fflush(stdout);
#endif
    while (!kbhit())
    {
	// Scan all the entries in the vocab table
	if (GetTicks() - starttime > waittime)
	{
	    count = 0;
	    ptr1 = vocab;
	    ptr1a = ptr2a = 0;
	    while (ptr1 < vocab+size)
	    {
		if ((ptr1[0] & TYPE_MASK) == TYPE_RELATED) ptr1 += 7;
		else if ((ptr1[0] & TYPE_MASK) == TYPE_RELATE2) ptr1 += 9;
		else if ((ptr1[0] & TYPE_MASK) == TYPE_EQUATE) ptr1 += 5;
		else if (ptr1[0] & DELETE_MASK) ptr1 += strlen(ptr1) + 3;
		else
		{
		    ptr2 = ptr1;
		    FirstTime = 1;
		    while ((ptr2 = GetNextItem(ptr2)) || FirstTime)
		    {
			FirstTime = 0;
			if (ptr2 == 0 && ptr1[0] != TYPE_NAME &&
			    ptr1[0] != TYPE_NOUN && ptr1[0] != TYPE_ADJECT)
			    break;
			if (ptr2 == 0 || ptr2[0] == TYPE_RELATED)
			{
			    count++;
			    if ((rand() % count) == 0)
			    {
				if (count == 1 || CheckMemory(ptr1, ptr2) == 0)
				{
				    ptr1a = ptr1;
				    ptr2a = ptr2;
				}
			    }
			}
		    }
		    ptr1 += strlen(ptr1) + 3;
		}
	    }
	    // Print the entry that was randomly selected
	    if (ptr1a && ptr2a)
	    {
		if (ptr2a[0] == TYPE_RELATED)
		{
		    putchar(8);
		    putchar(8);
		    PrintStatement(ptr1a+1, ptr2a, 1);
		    printf("> ");
#ifndef CSPIN_FLAG
                    fflush(stdout);
#endif
		}
	    }
	    else if (ptr1a)
	    {
	        putchar(8);
	        putchar(8);
		PrintQuestion(ptr1a);
	        printf("> ");
#ifndef CSPIN_FLAG
                fflush(stdout);
#endif
	    }
            waittime = ((rand() % 10) + 5) * TICKS_PER_SECOND;
            starttime = GetTicks();
	}
    }
}

// Save the vocab table
void SaveVocab(void)
{
#ifdef FILE_IO
    FILE *outfile = fopen("vocab.txt", "wb");
    if (outfile)
    {
        fwrite(vocab, 1, size, outfile);
        fclose(outfile);
	changeflag = 0;
	printf("< Saved\n");
    }
#else
    int *addr = &size;
    int i, val, val1;

    for (i = (size + 7) >> 2; i > 0; i--)
    {
	val = *addr;
	if (val != i2c_ReadLong(28, 0xa0, addr))
	{
	    i2c_WriteLong(28, 0xa0, addr, val);
	    while (i2c_WriteWait(28, 0xa0, addr));
	}
        addr += 4;
    }
#endif
}

int CheckCommand(char **tokens, int num, char *buffer)
{
    char *ptr1;
    int retval = 1;
    int caseshift = (tokens[0][0] >= 'A' && tokens[0][0] <= 'Z');

    if (num > 2) return 0;

    if (caseshift) tokens[0][0] += 'a' - 'A';

    if (strcmp(tokens[0], "help") == 0)
    {
        printf("$ Type a sentence, or enter one of these commands\n");
        printf("$ vocab [word] - Print the words I know\n");
        printf("$ dump         - Print the words as hex codes\n");
        printf("$ forget       - Remove a word from memroy\n");
        printf("$ save         - Save the words to EEPROM\n");
        printf("$ reset        - Reset the word table\n");
        printf("$ basic        - Reset the word table to a basic list\n");
	printf("$ history      - Print the short term memory\n");
        printf("$ goodbye      - Exit the program\n");
    }
    else if (strcmp(tokens[0], "goodbye") == 0)
    {
        if (changeflag)
        {
	    while (1)
	    {
	        printf("< Should I save?\n");
	        printf("> ");
	        gets(buffer);
	        ptr1 = SkipChar(buffer, ' ');
	        if (toupper(ptr1[0]) == 'Y')
	        {
	            SaveVocab();
		    break;
	        }
	        else if (toupper(ptr1[0]) == 'N') break;
	        printf("< I don't undertand\n");
	    }
        }
        printf("< Goodbye\n");
	exit(0);
    }
    else if (strcmp(tokens[0], "vocab") == 0)   DumpAsciiWords(tokens[1], num);
    else if (strcmp(tokens[0], "dump") == 0)    DumpHexWords();
    else if (strcmp(tokens[0], "forget") == 0)  DeleteWord(tokens,num);
    else if (strcmp(tokens[0], "save") == 0)    SaveVocab();
    else if (strcmp(tokens[0], "reset") == 0)   ResetVocab();
    else if (strcmp(tokens[0], "basic") == 0)   ResetBasicVocab();
    else if (strcmp(tokens[0], "history") == 0) PrintMemory();
    else if (strcmp(tokens[0], "debug") == 0)   debugmode = !debugmode;
    else                                        retval = 0;

    // Restore the case if needed
    if (caseshift) tokens[0][0] -= 'a' - 'A';

    return retval;
}

int main(void)
{
    int i;
    int num;
    char *ptr1;
    int ignore;
    int question;
    int caseshift;
    char buffer[100];
    char *tokens[50];
#ifdef FILE_IO
    FILE *infile = fopen("vocab.txt", "rb");

    if (infile)
    {
        size = fread(vocab, 1, BUF_SIZE, infile);
	fclose(infile);
    }

    if (size < BUF_SIZE) vocab[size] = 0;
#else
    i2c_Initialize(28);
#endif

    for (i = 0; i <= MEMORY_MASK; i++) Memory[i] = 0;

    //printf("< Hello\n");
    ptr1 = FindWords("you", "are");
    if (ptr1) PrintStatement("you", ptr1, 1);

    // Loop on each user input
    while (1)
    {
	WaitForInput();
	gets(buffer);
	num = tokenize(buffer, tokens);
	if (num == 0) continue;
	if (CheckCommand(tokens,num,buffer)) continue;

	// Check for a question mark at the end
        if (strcmp(tokens[num-1], "?") == 0)
        {
	    question = 1;
	    num--;
        }
	else
	    question = 0;

	// Check if all the words are in the vocab table
	ignore = 0;
	for (i = 0; i < num; i++)
	{
	    if (FindWord(tokens[i]) == 0)
	    {
		if(ignore = UnknownWord(tokens, num, i)) break;
	    }
	}
	if (ignore) continue;

	// Process the user input
        Process(tokens,num, question);
    }

    return 0;
}
/*
+-----------------------------------------------------------------------------+
|                          TERMS OF USE: MIT License                          |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
